/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * MySql Switch Resource - Andy Powell
 * 
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/options.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/cli.h>
#include <asterisk/module.h>
#include <asterisk/utils.h>
#include <asterisk/config.h>
#include <asterisk/config_pvt.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <mysql.h>
#include <asterisk/term.h>


static char *desc = "MySQL Switch module";

/* Move this shit to a config file you lazy ass mofo */
static char MYSQL_SERVER[80]="";
static char MYSQL_USER[80]="";
static char MYSQL_PWD[80]="";
static char MYSQL_DB[80]="";

static char MYSQL_TABLE[80]="dialplan";

#define LOOK_DONT_EXEC 0  /* Check for an application but don't run it */
#define LOOK_AND_EXEC 1   /* Check for an application and run it       */
#define ESCAPE_IT "'|_%'" /* escape for sql */

#define ADD_CONTEXTS 1
#define ADD_GLOBALS 2

#define BEFORE_SWITCH 0
#define AFTER_SWITCH 1

#define CONFIG_FILE "mysqlswitch.conf"


static int MySQL_generic_lookup(int dowhat, struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, int newstack, char *data);

STANDARD_LOCAL_USER;
LOCAL_USER_DECL;


static int add_includes(char *context, int when_to_add) 
{
	int res=0;
	int fnd=0;
        char sql[256]="";

        MYSQL mysql;
        MYSQL *mycon=NULL;
        MYSQL_RES *myres=NULL;
        MYSQL_ROW row;
				
        char realvalue[256];

        mysql_init(&mysql);
	mycon=mysql_real_connect(&mysql,MYSQL_SERVER,MYSQL_USER,MYSQL_PWD,MYSQL_DB,0,NULL,0);
	
	if (mycon) {
                snprintf(sql,sizeof(sql)-1,"SELECT includecontext FROM includes where targetcontext='%s' and includes.position=%d;",context,when_to_add);
                res=mysql_query(mycon,sql);
		
                if(res) {
			ast_log(LOG_ERROR,"Error in querying database, MYSQL ERROR:%s\n", mysql_error(&mysql));
			mysql_close(mycon);
			res=-1;
		} else {
			myres=mysql_use_result(mycon);
                        if(!myres) {
	                        ast_log(LOG_ERROR,"MYSQL Error:%s\n",mysql_error(&mysql));
		                mysql_close(mycon);
		                res=-1;
			} else {			
        	                while ((row=mysql_fetch_row(myres))) {
	        	                if(row) {
						ast_context_add_include(context, row[0], desc);
						res=0;
					}
						
				}

                	        mysql_free_result(myres);

				mysql_close(mycon);
				res=0;
			}
		}
	} else {
	        ast_log(LOG_ERROR, "Could not connect to database server %s\n",MYSQL_SERVER);
	        res=-1;
	}

	return res;
}

static int remove_globals() 
{
	int res=0;
	int fnd=0;
        char sql[256]="";

        MYSQL mysql;
        MYSQL *mycon=NULL;
        MYSQL_RES *myres=NULL;
        MYSQL_ROW row;
				
        char realvalue[256];

        mysql_init(&mysql);
	mycon=mysql_real_connect(&mysql,MYSQL_SERVER,MYSQL_USER,MYSQL_PWD,MYSQL_DB,0,NULL,0);
	
	if (mycon) {
		strncpy(sql,"SELECT name from globals;", sizeof(sql)-1);
                res=mysql_query(mycon,sql);
		
                if(res) {
			ast_log(LOG_ERROR,"Error in querying database, MYSQL ERROR:%s\n", mysql_error(&mysql));
			mysql_close(mycon);
			res=-1;
		} else {
			myres=mysql_use_result(mycon);
                        if(!myres) {
	                        ast_log(LOG_ERROR,"MYSQL Error:%s\n",mysql_error(&mysql));
		                mysql_close(mycon);
		                res=-1;
			} else {			
        	                while ((row=mysql_fetch_row(myres))) {
	        	                if(row) {
						memset(realvalue, 0, sizeof(realvalue));
						pbx_builtin_setvar_helper(NULL, row[0], realvalue);
						}
						
						fnd=1;
					
				}

				if (fnd >0)
	                	        mysql_free_result(myres);

				mysql_close(mycon);
				res=0;
			}
		}
	} else {
	        ast_log(LOG_WARNING, "Could not connect to database server %s\n",MYSQL_SERVER);
	        res=-1;
	}

	return res;
}


static int add_globals_contexts(int itemtype) 
{
	int res=0;
	int fnd=0;
        char sql[256]="";

        MYSQL mysql;
        MYSQL *mycon=NULL;
        MYSQL_RES *myres=NULL;
        MYSQL_ROW row;
				
        char realvalue[256];

        mysql_init(&mysql);
	mycon=mysql_real_connect(&mysql,MYSQL_SERVER,MYSQL_USER,MYSQL_PWD,MYSQL_DB,0,NULL,0);
	
	if (mycon) {
		if (itemtype == ADD_GLOBALS) {
			strncpy(sql,"SELECT name,value from globals where active=1;", sizeof(sql)-1);
		} else {
	                strncpy(sql,"SELECT distinct name,switch from contexts;", sizeof(sql)-1);
					
		}
                res=mysql_query(mycon,sql);
		
                if(res) {
			ast_log(LOG_ERROR,"Error in querying database, MYSQL ERROR:%s\n", mysql_error(&mysql));
			mysql_close(mycon);
			res=-1;
		} else {
			myres=mysql_use_result(mycon);
                        if(!myres) {
	                        ast_log(LOG_ERROR,"MYSQL Error:%s\n",mysql_error(&mysql));
		                mysql_close(mycon);
		                res=-1;
			} else {			
        	                while ((row=mysql_fetch_row(myres))) {
	        	                if(row) {
						if (itemtype == ADD_GLOBALS) {				
							memset(realvalue, 0, sizeof(realvalue));
							pbx_substitute_variables_helper(NULL, row[1], realvalue, sizeof(realvalue) - 1);
							pbx_builtin_setvar_helper(NULL, row[0], realvalue);
						} else {
							if ((ast_context_find(row[0])) == NULL)
		                                                ast_context_create(NULL, row[0], desc);
        	                                        // add switch here too

							add_includes(row[0],BEFORE_SWITCH);
							
							if (atoi(row[1]) ==1) {
                	                                	ast_context_add_switch(row[0], "MySQLSwitch", NULL, desc);
							}
							/* Do we want to add them even if there is no switch, but the user
							specified the insertion after the switch? We should be told. At the
							moment we will check/add */ 
							add_includes(row[0],AFTER_SWITCH);
						}
						
						fnd=1;
					}
				}

                	        mysql_free_result(myres);

				mysql_close(mycon);
				res=0;
			}
		}
	} else {
	        ast_log(LOG_ERROR, "Could not connect to database server %s\n",MYSQL_SERVER);
	        res=-1;
	}

	return res;
}


static int MySQLSwitch_exists(struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, char *data)
{
	int res = 0; 

	res=MySQL_generic_lookup(LOOK_DONT_EXEC, chan, context, exten, priority, callerid, 0, data);

	return res;
}

static int SQLiteSwitch_canmatch(struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, char *data)
{

	int res = 1; 
	return res;

}
static int MySQLSwitch_exec(struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, int newstack, char *data)
{
        int res = 0;

        res=MySQL_generic_lookup(LOOK_AND_EXEC, chan, context, exten, priority, callerid, newstack, data);

        return res;
			

}

static int MySQL_generic_lookup(int dowhat, struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, int newstack, char *data)
{
	struct ast_app *app;
	int res = 0;
	//time_t now;
	char app_data[1024]="";
	char tmp_data[1024]="";
	char app_name[256]="";
	char tmp2[1024]="";
	char tmp[80]="";
	char tmp3[80]="";
	char sql[2048]="";
	char tapp[80]="";
	int foundit=0;

	MYSQL mysql;
	MYSQL *mycon=NULL;
	MYSQL_RES *myres;
        MYSQL_ROW row;

	//time(&now);

	memset(&app_data,0,sizeof(app_data));
	memset(&tmp_data,0,sizeof(tmp_data));
	
	// Connect to db
	mysql_init(&mysql);
        mycon=mysql_real_connect(&mysql,MYSQL_SERVER,MYSQL_USER,MYSQL_PWD,MYSQL_DB,0,NULL,0);
		
	if (mycon) {
		snprintf(sql,sizeof(sql)-1,"SELECT app,data FROM dialplan WHERE context='%s' AND pri=%d AND exten='%s' LIMIT 1;",context,priority,exten);
        	res=mysql_query(mycon,sql);
		
		if(res) {
			ast_log(LOG_ERROR,"Error in querying database, MYSQL ERROR:%s\n", mysql_error(&mysql));
                        mysql_close(mycon);
			res=-1;
												
		} else {
                        myres=mysql_use_result(mycon);

			if(!myres) {
				ast_log(LOG_ERROR,"MYSQL Error:%s\n",mysql_error(&mysql));
                                res=-1;
		                mysql_close(mycon);
								
			} else {
				row=mysql_fetch_row(myres);
				
				if(row) {
					snprintf (app_name, sizeof (app_name)-1, "%s\0",row[0]);
					snprintf (tmp_data, sizeof (tmp_data)-1, "%s\0",row[1]);
					mysql_free_result(myres);
					mysql_close(mycon);
					
					if (dowhat == LOOK_AND_EXEC) {
						// Check app, parse vars, check 
						app = pbx_findapp(app_name);
						if (app) {
							pbx_substitute_variables_helper(chan, tmp_data, app_data, sizeof(app_data) - 1);
		                                	if (option_verbose > 10)
			                                                ast_verbose( VERBOSE_PREFIX_3 "Executing %s(%s) on channel %s\n",
	                                                                term_color(tmp, app_name, COLOR_BRCYAN, 0, sizeof(tmp)),
									term_color(tmp2, app_data, COLOR_BRMAGENTA, 0, sizeof(tmp2)),
									term_color(tmp3, chan->name, COLOR_YELLOW, 0, sizeof(tmp3)));
									
//							ast_log(LOG_ERROR,"Found app, %s, running it\n",app_name);
							res = pbx_exec(chan, app,app_data,1);
						} else {
							ast_log(LOG_WARNING, "application %s not registered\n",row[0]);
				                        res = -1;
						}			
					} else {
						res=-1;
						snprintf (app_name, sizeof (app_name)-1, "%s\0",row[0]);
			                        app = pbx_findapp(app_name);
			                        if (app) { 
//							ast_log(LOG_ERROR,"Found app, %s but we're only looking\n",app_name);
							res=1;
						}												
						
					}
				} else {  //PATTERN Match
					if (strlen(exten) >1) {
						/* Initial SQL must have returned no results */
						/* so now we need to try a pattern match     */
						mysql_free_result(myres);
						
			        	        snprintf(sql,sizeof(sql)-1,"SELECT exten,app,data FROM dialplan WHERE context='%s' AND pri=%d AND exten like %s ESCAPE '|' ORDER BY LENGTH(exten) DESC;",context,priority,ESCAPE_IT);
						res=mysql_query(mycon,sql);
						
						if(res) {
				                        ast_log(LOG_ERROR,"Error in querying database, MYSQL ERROR:%s\n", mysql_error(&mysql));
			                        	mysql_close(mycon);
							res=-1;
						} else {
			                        	myres=mysql_use_result(mycon);
				                        if(!myres) {
				                                ast_log(LOG_ERROR,"MYSQL Error:%s\n",mysql_error(&mysql));
					                        mysql_close(mycon);
				                                res=-1;
				                        } else {
								res=-1;
								/* searching for pattern match */
								foundit=0;
								
								while ((row=mysql_fetch_row(myres))) {	
									if (row) {
										snprintf(tmp_data,sizeof(tmp_data)-1,"%s\0", row[0]);
										strncpy(tmp_data,"_",1);
										
										foundit = ast_extension_match(tmp_data, exten);
										if (foundit == 1) {
//											ast_log(LOG_NOTICE,"Someone is suggesting %s matches %s (%d)\n",tmp_data,exten,foundit);
											snprintf (app_name, sizeof (app_name)-1, "%s\0",row[1]);
											snprintf (tmp_data, sizeof (tmp_data)-1, "%s\0",row[2]);
											foundit = 4;
											break;
										}
									} else {
										break;
									}
						    		}	
                                                                mysql_free_result(myres);
			                		        mysql_close(mycon);
						
						    		if (foundit == 4) {
				                        		if (dowhat == LOOK_AND_EXEC) {
        	                        				        app = pbx_findapp(app_name);
					                                        if (app) {
					                                                pbx_substitute_variables_helper(chan, tmp_data, app_data, sizeof(app_data) - 1);

	                               				                        if (option_verbose > 10)
       		                                                       			        ast_verbose( VERBOSE_PREFIX_3 "Executing %s(%s) on channel %s\n",
				                                                                        term_color(tmp, app_name, COLOR_BRCYAN, 0, sizeof(tmp)),
                       					                                                term_color(tmp2, app_data, COLOR_BRMAGENTA, 0, sizeof(tmp2)),
                                		               			                        term_color(tmp3, chan->name, COLOR_YELLOW, 0, sizeof(tmp3)));
												
				        		                                res = pbx_exec(chan, app,app_data,1);
                                       					        } else {
			                                                        	ast_log(LOG_WARNING, "application %s not registered\n",app_name);
                        	       				                        res = -1;
			        	                                        }
                               					        } else {
			                                                	res=1;
			                                        	}
						    		} else {
						    			res=-1;
						    		}
							}	
						}
					
					} else {
						mysql_free_result(myres);
						mysql_close(mycon);
					}												
				}
			}
		}

	} else {
		ast_log(LOG_ERROR, "Could not connect to database server %s\n",MYSQL_SERVER);
		res=-1;
	}
	
	return res;
}


static int MySQLSwitch_matchmore(struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, char *data)
{

	int res = 0; 
	return res;
}

static int MySQLSwitch_canmatch(struct ast_channel *chan, char *context, char *exten, int priority, char *callerid, char *data)
{

        int res = 1;
        return res;
}


static struct ast_switch mysql_switch = 
	{
		name: 			"MySQLSwitch",
		description:		"Read the dialplan from a MySQL database",
		exists:			MySQLSwitch_exists,
		canmatch:		MySQLSwitch_canmatch,
		exec:			MySQLSwitch_exec,
		matchmore:		MySQLSwitch_matchmore,
	};


static int load_config(int readd)
{
      struct ast_config *cfg;
      struct ast_variable *v;

      cfg = ast_load(CONFIG_FILE);

	if (readd == 1) {
	        ast_verbose(VERBOSE_PREFIX_2 "Removing global variables\n");
	        remove_globals();

        	ast_verbose(VERBOSE_PREFIX_2 "Remove contexts and switching\n");
	        ast_context_destroy(NULL, desc);
	}				


      if (!cfg)
                {
            ast_log(LOG_NOTICE, "Unable to load config %s, nasty shit will happen!\n", CONFIG_FILE);
            return 0;
              }

                strncpy(MYSQL_SERVER, "127.0.0.1", sizeof(MYSQL_SERVER) - 1);
                strncpy(MYSQL_USER, "root", sizeof(MYSQL_USER) - 1);
                strncpy(MYSQL_PWD, "createaconfigumuppet", sizeof(MYSQL_PWD) - 1);
                strncpy(MYSQL_DB, "dbname", sizeof(MYSQL_DB) - 1);
                strncpy(MYSQL_TABLE, "table", sizeof(MYSQL_TABLE) - 1);

              v = ast_variable_browse(cfg, "general");

              while(v)
                        {

                        if (!strcasecmp(v->name, "mysql_server"))
                                {
                                strncpy(MYSQL_SERVER, v->value, sizeof(MYSQL_SERVER)-1);
                                }
                        else if (!strcasecmp(v->name, "mysql_user"))
                                {
                                strncpy(MYSQL_USER, v->value, sizeof(MYSQL_USER)-1);
                                }
                        else if (!strcasecmp(v->name, "mysql_pwd"))
                                {
                                strncpy(MYSQL_PWD, v->value, sizeof(MYSQL_PWD)-1);
                                }
                        else if (!strcasecmp(v->name, "mysql_db"))
                                {
                                strncpy(MYSQL_DB, v->value, sizeof(MYSQL_DB)-1);
                                }

                    v = v->next;
                }

                ast_destroy(cfg);

	        if (readd == 1) {
	        	ast_verbose(VERBOSE_PREFIX_2 "Adding global variables\n");
		        add_globals_contexts(ADD_GLOBALS);
  		        ast_verbose(VERBOSE_PREFIX_2 "Adding contexts and switching\n");
		        add_globals_contexts(ADD_CONTEXTS);
		}			
                return 0;

}



int reload(void) 
{
	
	return load_config(1);

}

int load_module(void)
{
	int res = 0;

	load_config(0);

	if (ast_register_switch(&mysql_switch))
		ast_log(LOG_ERROR, "Unable to register MySQL switch\n");
	else {
		ast_verbose(VERBOSE_PREFIX_2 "Registered MySQL switch\n");
                ast_verbose(VERBOSE_PREFIX_2 "Adding global variables\n");
                add_globals_contexts(ADD_GLOBALS);
		ast_verbose(VERBOSE_PREFIX_2 "Adding contexts and switching\n");
		add_globals_contexts(ADD_CONTEXTS);
	}

	return 0;
}


int unload_module(void)
{
	int res = 0;

	ast_verbose(VERBOSE_PREFIX_2 "MySQL Switch Disabled\n");
	ast_unregister_switch(&mysql_switch);
	
        ast_verbose(VERBOSE_PREFIX_2 "Removing global variables\n");
        remove_globals();

	ast_verbose(VERBOSE_PREFIX_2 "Remove contexts and switching\n");
	ast_context_destroy(NULL, desc);							


	return res;
}

char *description(void)
{
	return desc;
}

int usecount(void)
{
	int res=0;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}



